#@IgnoreInspection BashAddShebang
# String functionality (Koalephant Shell Script Library)
# Version: PACKAGE_VERSION
# Copyright: 2014-2017, Koalephant Co., Ltd
# Author: Stephen Reay <stephen@koalephant.com>, Koalephant Packaging Team <packages@koalephant.com>

# Get a line from a multiline variable
#
# Input:
# $1 - the line number to fetch
# $2...n - all remaining arguments are treated as the text to search
#
# Output:
# The text on the specified line
#
k_string_get_line() {
	# shellcheck disable=SC2039
	local lineNo

	k_requires_args k_string_get_line $#

	lineNo="$(k_int_parse_error "$1")" || return 1
	shift

	printf -- '%s\n' "$@" | sed -E -n -e "${lineNo}p"
}

# Right Pad a string to the given length
#
# Input:
# $1 - the length to pad to
# $n - all remaining arguments are treated as the text to pad

# Output:
# The string, with padding applied
k_string_pad_right() {
	# shellcheck disable=SC2039
	local length

	k_requires_args k_string_pad_right $#

	length="$(k_int_parse_error "$1")" || return 1
	shift

	printf -- "%-${length}s" "$*"
}

# Left Pad a string to the given length
#
# Input:
# $1 - the length to pad to
# $n - all remaining arguments are treated as the text to pad
# PAD_CHAR - if set, the character to pad with (defaults to space)

# Output:
# The string, with padding applied
k_string_pad_left() {
	# shellcheck disable=SC2039
	local length

	k_requires_args k_string_pad_left $#

	length="$(k_int_parse_error "$1")" || return 1
	shift

	printf -- "%+${length}s" "$*"
}

# Convert a string to lower case
#
# Input:
# $1...n - all arguments are treated as the text to convert
#
# Output:
# The string, converted to lower case
k_string_lower() {
	printf -- '%s\n' "$*" | LC_CTYPE=C tr '[:upper:]' '[:lower:]'
}

# Convert a string to upper case
#
# Input:
# $1...n - all arguments are treated as the text to convert
#
# Output:
# The string, converted to upper case
k_string_upper() {
	printf -- '%s\n' "$*" | LC_CTYPE=C tr '[:lower:]' '[:upper:]'
}

# Test if a string contains the given substring
#
# Input:
# $1 - the substring to check for
# $2...n - all arguments are treated as the string to check in
#
# Return:
# 0 if substring is found, 1 if not
k_string_contains() {
	# shellcheck disable=SC2039
	local substring string

	k_requires_args k_string_contains $# 2

	substring="${1}"
	shift
	string="$*"

	if [ "${#string}" -gt 0 ] && [ "${string%*${substring}*}" != "${string}" ]; then
		return 0
	fi

	return 1
}

# Test if a string starts with the given substring
#
# Input:
# $1 - the substring to check for
# $2...n - all arguments are treated as the string to check in
#
# Return:
# 0 if the string starts with substring, 1 if not
k_string_starts_with() {
	# shellcheck disable=SC2039
	local substring string

	k_requires_args k_string_starts_with $# 2

	substring="${1}"
	shift
	string="$*"

	if [ "${#string}" -gt 0 ] && [ "${string%%${substring}*}" = "" ]; then
		return 0
	fi

	return 1
}

# Test if a string ends with the given substring
#
# Input:
# $1 - the substring to check for
# $2...n - all arguments are treated as the string to check in
#
# Return:
# 0 if the string ends with substring, 1 if not
k_string_ends_with() {
	# shellcheck disable=SC2039
	local substring string

	k_requires_args k_string_ends_with $# 2

	substring="${1}"
	shift
	string="$*"

	if [ "${#string}" -gt 0 ] && [ "${string##*${substring}}" = "" ]; then
		return 0
	fi

	return 1
}

# Remove a substring from the start of a string
#
# Input:
# $1 - the substring to remove
# $2...n - all arguments are treated as the string to operate on
# $REMOVE_GREEDY flag to enable greedy removal if the substring contains the '*' character
#
# Output:
# The string with substring removed from the start
k_string_remove_start() {
	# shellcheck disable=SC2039
	local substring string

	k_requires_args k_string_remove_start $# 2

	substring="${1}"
	shift
	string="$*"

	if [ "${REMOVE_GREEDY:-false}" = true ]; then
		printf -- '%s\n' "${string##${substring}}"
	else
		printf -- '%s\n' "${string#${substring}}"
	fi
}

# Remove a substring from the end of a string
#
# Input:
# $1 - the substring to remove
# $2...n - all arguments are treated as the string to operate on
# $REMOVE_GREEDY flag to enable greedy removal if the substring contains the '*' character
#
# Output:
# The string with substring removed from the end
k_string_remove_end() {
	# shellcheck disable=SC2039
	local substring string

	k_requires_args k_string_remove_end $# 2

	substring="${1}"
	shift
	string="$*"

	if [ "${REMOVE_GREEDY:-false}" = true ]; then
		printf -- '%s\n' "${string%%${substring}}"
	else
		printf -- '%s\n' "${string%${substring}}"
	fi

}

# Join strings with an optional separator
#
# Input:
# $1 - the separator string
# $2...n - the strings to join
#
# Output:
# the strings joined by the separator
#
# Example:
# k_string_join "-" this is a nice day # "this-is-a-nice-day"
k_string_join() {
	# shellcheck disable=SC2039
	local separator

	separator="${1-}"
	shift
	if [ $# -eq 1 ]; then
		printf -- '%s' "$1"
	elif [ $# -gt 1 ]; then
		printf -- '%s' "$1"
		shift
		printf -- "${separator}%s" "$@"
	fi
}

# Remove leading and trailing whitespace from strings
#
# Input:
# $1...n - the strings to trim
#
# Output:
# the strings with leading/trailing whitespace removed
k_string_trim() {
	printf -- '%s\n' "$*" | sed -e "s/^[[:space:]]*//g; s/[[:space:]]*$//g;"
}

# Check if the given keyword is in the list of valid keywords, or error
#
# Input:
# $1 - the input to check
# $2...n - the list of valid keywords
#
# Output:
# the keyword if valid
#
# Return:
# 0 if keyword is valid, 1 if not
k_string_keyword_error() {
	# shellcheck disable=SC2039
	local input
	k_requires_args k_string_keyword_error $# 2

	input="${1}"
	shift
	while [ $# -gt 0 ]; do
		if [ "${input}" = "${1}" ]; then
			printf -- '%s\n' "$1"
			return 0
		fi
		shift
	done

	return 1
}

# Get the selected keyword from a list, or default
#
# Input:
# $1 - the input to check
# $2...n - the list of valid keywords. The last value is used as the default if none match
#
# Output:
# the valid keyword or default
k_string_keyword() {
	# shellcheck disable=SC2039
	local value

	k_requires_args k_string_keyword $# 2

	if value="$(k_string_keyword_error "$@")"; then
		printf -- '%s\n' "${value}"
	else
		shift $(($#-1))
		printf -- '%s\n' "$1"
	fi
}
