#!/bin/sh

. ../base.lib.sh
. ../environment.lib.sh
. ../bool.lib.sh
. ../number.lib.sh
. ../fs.lib.sh
. ../string.lib.sh
. ../hash.lib.sh


worker_k_hash_algos() {
	local algos expected="$1"

	algos="$(k_hash_algos)"
	assertTrue "(hash module=${KOALEPHANT_HASH_MODULE}) Listing algos returns success" $?
	assertEquals "(hash module=${KOALEPHANT_HASH_MODULE}) Known hash algos are available" "${expected}" "${algos}"
}

test_k_hash_algos (){
	local phpAlgos shasumAlgos algosumAlgos

	phpAlgos="$(php -r 'echo implode(PHP_EOL, hash_algos());' | sort -u)"
	shasumAlgos="$(printf -- '%s\n' sha1 sha224 sha256 sha384 sha512 sha512224 sha512256 | sort -u)"
	algosumAlgos="$(printf -- '%s\n' blake2 md5 sha1 sha224 sha256 sha384 sha512 | sort -u)"

	unset KOALEPHANT_HASH_MODULE
	worker_k_hash_algos "$(printf -- '%s\n' "${phpAlgos}" "${shasumAlgos}" "${algosumAlgos}" | sort -u)"

	KOALEPHANT_HASH_MODULE=php worker_k_hash_algos "${phpAlgos}"
	KOALEPHANT_HASH_MODULE=shasum worker_k_hash_algos "${shasumAlgos}"
	KOALEPHANT_HASH_MODULE=algosum worker_k_hash_algos "${algosumAlgos}"
}


test_k_hash_modules (){
	local modules expected

	expected="$(printf '%s\n' php shasum algosum | sort)"
	modules="$(k_hash_modules)"
	assertTrue 'Listing hash modules returns success' $?
	assertEquals 'Known hash modules are loaded' "${expected}" "${modules}"
}

worker_k_hash_generate() {
	local algo string hash generated tempFile
	KOALEPHANT_HASH_MODULE="$1"

	while IFS=';' read -r algo string hash; do
		if ! k_hash_op avail "${algo}" 2>/dev/null; then
			startSkipping
		fi
		generated="$(printf -- '%s' "${string}" | k_hash_generate "${algo}" - 2>/dev/null)"
		assertTrue "(hash module=${KOALEPHANT_HASH_MODULE}) Generating ${algo} hash for string '${string}' returns success" $?
		assertEquals "(hash module=${KOALEPHANT_HASH_MODULE}) ${algo} Hash for '$string' generated correctly" "${hash}" "${generated}"
		if isSkipping; then
			endSkipping
		fi
	done < ./hashes.txt

	while IFS=';' read -r algo string hash; do
		if ! k_hash_op avail "${algo}" 2>/dev/null; then
			startSkipping
		fi
		tempFile="$(mktemp)"
		printf -- '%s' "${string}" > "${tempFile}"
		generated="$(k_hash_generate "${algo}" "${tempFile}" 2>/dev/null)"
		assertTrue "(hash module=${KOALEPHANT_HASH_MODULE}) Generating ${algo} hash for contents of '${tempFile}' returns success" $?
		assertEquals "(hash module=${KOALEPHANT_HASH_MODULE}) ${algo} Hash for contents of '${tempFile}' generated correctly" "${hash}" "${generated}"
		rm -f "${tempFile}"
		if isSkipping; then
			endSkipping
		fi
	done < ./hashes.txt
}

test_k_hash_generate (){
	worker_k_hash_generate ''
	worker_k_hash_generate php
	worker_k_hash_generate shasum
	worker_k_hash_generate algosum
}

worker_k_hash_verify() {
	local algo string hash generated tempFile
	KOALEPHANT_HASH_MODULE="$1"

	while IFS=';' read -r algo string hash; do
		if ! k_hash_op avail "${algo}" 2>/dev/null; then
			startSkipping
		fi
		printf -- '%s' "${string}" | k_hash_verify "${algo}" - "${hash}" 2>/dev/null
		assertTrue "(hash module=${KOALEPHANT_HASH_MODULE}) Verifying ${algo} hash for string '${string}' returns success" $?
		if isSkipping; then
			endSkipping
		fi
	done < ./hashes.txt

	while IFS=';' read -r algo string hash; do
		if ! k_hash_op avail "${algo}" 2>/dev/null; then
			startSkipping
		fi
		tempFile="$(mktemp)"
		printf -- '%s' "${string}" > "${tempFile}"
		k_hash_verify "${algo}" "${tempFile}" "${hash}" 2>/dev/null
		assertTrue "(hash module=${KOALEPHANT_HASH_MODULE}) Verifying ${algo} hash for contents of '${tempFile}' returns success" $?
		rm -f "${tempFile}"
		if isSkipping; then
			endSkipping
		fi
	done < ./hashes.txt
}

test_k_hash_verify (){
	worker_k_hash_verify ''
	worker_k_hash_verify php
	worker_k_hash_verify shasum
	worker_k_hash_verify algosum
}

test_k_hash_op (){

	KOALEPHANT_HASH_MODULE='none'
	output="$(k_hash_op 'whoami' 2>&1 1>/dev/null)"
	assertFalse 'Return is error when no hash modules can run' $?
	assertEquals 'Output is error message when no hash modules can run' 'No hash module available' "${output}"

	KOALEPHANT_HASH_MODULE=php

	output="$(k_hash_op 'whoami')"
	assertTrue 'Return is success when php hash module can run' $?
	assertEquals 'Output is php Version info' "$(php --version)" "${output}"


	KOALEPHANT_HASH_MODULE=shasum
	output="$(k_hash_op 'whoami' '')"
	assertTrue 'Return is success when shasum hash module can run' $?
	assertEquals 'Output is shasum Version info' "$(shasum --version)" "${output}"

	KOALEPHANT_HASH_MODULE=algosum
	output="$(k_hash_op 'whoami' '')"
	assertTrue 'Return is success when algosum hash module can run' $?
	assertEquals 'Output is algosum Version info' "$(sha1sum --version)" "${output}"
}
